from typing import TYPE_CHECKING, Self

if TYPE_CHECKING:
    from wame.engine import Engine

from wame.vector import IntVector2
from wame.button import Button

import pygame

class Scene:
    '''Handles all events and rendering for the engine'''

    def __init__(self, engine:'Engine') -> None:
        '''
        Instantiate a new scene
        
        Parameters
        ----------
        engine : `wame.Engine`
            The engine instance
        '''
        
        self.engine:'Engine' = engine
        '''The engine running the scene'''

        self._buttons:list[Button] = []
    
    def _check_events(self) -> None:
        for event in pygame.event.get():
            match event.type:
                case pygame.KEYDOWN:
                    self.on_key_pressed(event.key)
                case pygame.KEYUP:
                    self.on_key_released(event.key)
                case pygame.MOUSEBUTTONDOWN:
                    mousePosition:IntVector2 = IntVector2.from_tuple(event.pos)

                    for button in self._buttons:
                        button._check_click(mousePosition)

                    self.on_mouse_pressed(mousePosition)
                case pygame.MOUSEBUTTONUP:
                    mousePosition:IntVector2 = IntVector2.from_tuple(event.pos)

                    self.on_mouse_released(mousePosition)
                case pygame.MOUSEMOTION:
                    mousePosition:IntVector2 = IntVector2.from_tuple(event.pos)

                    for button in self._buttons:
                        button._check_hover(mousePosition)

                    self.on_mouse_move(mousePosition)
                case pygame.MOUSEWHEEL:
                    mousePosition:IntVector2 = IntVector2.from_tuple(event.pos)

                    self.on_mouse_wheel_scroll(mousePosition, event.y)
                case pygame.QUIT:
                    self.engine._running = False

                    self.on_quit()
                case pygame.WINDOWENTER:
                    self.engine._set_fps = self.engine.settings.max_fps
                case pygame.WINDOWLEAVE:
                    self.engine._set_fps = self.engine.settings.tabbed_fps

    def _check_keys(self) -> None:
        keys:pygame.key.ScancodeWrapper = pygame.key.get_pressed()
        pressed:list[int] = [key for key, is_pressed in enumerate(keys) if is_pressed]

        for key in pressed:
            self.on_key_pressing(key)
    
    def _cleanup(self) -> None:
        self.on_cleanup()
    
    def _first(self) -> Self:
        self.on_first()

        return self

    def _update(self) -> None:
        self.engine._screen.fill((0, 0, 0))

        self.on_update()
        
        pygame.display.flip()
        self.engine._deltaTime = self.engine._clock.tick(self.engine._set_fps) / 1000.0
    
    def on_cleanup(self) -> None:
        '''
        Code below should be executed when the scene is being switched/cleaned up

        Example
        -------
        ```python
        # Setup
        class MyScene(wame.Scene):
            def __init__(self, engine) -> None:
                super().__init__(engine)
            
            def on_cleanup(self) -> None:
                ... # Terminate background threads, save data, etc.
        ```
        '''

        ...

    def on_first(self) -> None:
        '''
        Code below should be executed when the scene is about to be rendered

        Example
        -------
        ```python
        # Setup
        class MyScene(wame.Scene):
            def __init__(self, engine) -> None:
                super().__init__(engine)
            
            def on_first(self) -> None:
                ... # Start game timers, etc.
        ```
        '''

        ...

    def on_key_pressed(self, key:int) -> None:
        '''
        Code below should be executed when a key is pressed

        Example
        -------
        ```python
        # Setup
        class MyScene(wame.Scene):
            def __init__(self, engine) -> None:
                super().__init__(engine)
            
            def on_key_pressed(self, key:int) -> None:
                ... # Pause game, display UI, etc.
        ```
        '''
        
        ...
    
    def on_key_pressing(self, key:int) -> None:
        '''
        Code below should be executed when a key is being pressed

        Example
        -------
        ```python
        # Setup
        class MyScene(wame.Scene):
            def __init__(self, engine) -> None:
                super().__init__(engine)
            
            def on_key_pressing(self, key:int) -> None:
                ... # Move forward, honk horn, etc.
        ```
        '''
        
        ...
    
    def on_key_released(self, key:int) -> None:
        '''
        Code below should be executed when a key is released

        Example
        -------
        ```python
        # Setup
        class MyScene(wame.Scene):
            def __init__(self, engine) -> None:
                super().__init__(engine)
            
            def on_key_released(self, key:int) -> None:
                ... # Stop moving forward, etc.
        ```
        '''
        
        ...
    
    def on_mouse_move(self, mousePosition:IntVector2) -> None:
        '''
        Code below should be executed when the mouse moves

        Example
        -------
        ```python
        # Setup
        class MyScene(wame.Scene):
            def __init__(self, engine) -> None:
                super().__init__(engine)
            
            def on_mouse_move(self, mousePosition:wame.IntVector2) -> None:
                ... # Rotate character, etc.
        ```
        '''
        
        ...
    
    def on_mouse_pressed(self, mousePosition:IntVector2) -> None:
        '''
        Code below should be executed when a mouse button was pressed

        Example
        -------
        ```python
        # Setup
        class MyScene(wame.Scene):
            def __init__(self, engine) -> None:
                super().__init__(engine)
            
            def on_mouse_pressed(self, mousePosition:wame.IntVector2) -> None:
                ... # Start shooting, rotate character, etc.
        ```
        '''
        
        ...
    
    def on_mouse_released(self, mousePosition:IntVector2) -> None:
        '''
        Code below should be executed when a mouse button was released

        Example
        -------
        ```python
        # Setup
        class MyScene(wame.Scene):
            def __init__(self, engine) -> None:
                super().__init__(engine)
            
            def on_mouse_released(self, mousePosition:wame.IntVector2) -> None:
                ... # Shoot arrow, stop shooting, etc.
        ```
        '''
        
        ...
    
    def on_mouse_wheel_scroll(self, mousePosition:IntVector2, amount:int) -> None:
        '''
        Code below should be executed when the scroll wheel moves

        Example
        -------
        ```python
        # Setup
        class MyScene(wame.Scene):
            def __init__(self, engine) -> None:
                super().__init__(engine)
            
            def on_mouse_wheel_scroll(self, mousePosition:wame.IntVector2, amount:int) -> None:
                if amount > 0:
                    print(f"Scroll wheel moved up @ {mousePosition}!")
                else:
                    print(f"Scroll wheel moved down @ {mousePosition}!")
        ```
        '''

        ...

    def on_start(self) -> None:
        '''
        Code below should be executed when the engine is starting

        Example
        -------
        ```python
        # Setup
        class MyScene(wame.Scene):
            def __init__(self, engine) -> None:
                super().__init__(engine)
            
            def on_start(self) -> None:
                ... # Create threads, background processes for networking, etc.
        ```
        '''
        
        ...
    
    def on_update(self) -> None:
        '''
        Code below should be executed every frame

        Example
        -------
        ```python
        # Setup
        class MyScene(wame.Scene):
            def __init__(self, engine) -> None:
                super().__init__(engine)
            
            def on_update(self) -> None:
                ... # Render text, objects, etc.
        ```
        '''
        
        ...
    
    def on_quit(self) -> None:
        '''
        Code below should be executed when the engine quits

        Example
        -------
        ```python
        # Setup
        class MyScene(wame.Scene):
            def __init__(self, engine) -> None:
                super().__init__(engine)
            
            def on_quit(self) -> None:
                ... # Save data, cleanup objects, etc.
        ```
        '''
        
        ...